#!/usr/bin/env python
############################################################################
# Joshua R. Boverhof, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################
import sys, time, urlparse, optparse
from twisted.python import log
from twisted.internet import reactor
from ConfigParser import ConfigParser
import pyGridWare.generated.stubs.Counter_services as COUNTER
from pyGridWare.notification.NotificationConsumer import NotificationConsumer
from twisted.python import threadable
threadable.init()


###########################################################################
# GLOBALS
###########################################################################
DIALECT = "http://docs.oasis-open.org/wsn/2004/06/TopicExpression/Simple"
TEST = 'CounterService'
CONFIG_FILE = 'config.txt'
SEC_PATHS = 'Paths'
SEC_LOCATION = 'Location'

CP = ConfigParser()
CP.read(CONFIG_FILE)
HOST = CP.get(SEC_LOCATION, 'host')
PORT = CP.getint(SEC_LOCATION, 'port')
SSL = CP.getint('security', 'ssl')
DEBUG = CP.getint('logging', 'debug')


###########################################################################
# OPTIONS
###########################################################################
op = optparse.OptionParser()
op.add_option("-s", "--host",
              action="store", dest="host", default=HOST,
              help="location of service container")
op.add_option("-p", "--port",
              action="store", dest="port", type="int", default=PORT,
              help="port of service container")
op.add_option("-t", "--ssl",
              action="store", dest="ssl", type="int", default=SSL,
              help="use secure transport")
op.add_option("-d", "--debug",
              action="store", dest="debug", type="int", default=DEBUG,
              help="debug")
(OPTIONS, ARGS) = op.parse_args()



class TopicExpression:
    '''
    '''
    def __init__(self):
        self._attrs = {}
        self._attrs['Dialect'] = DIALECT

    def setMixed(self, value):
        self._text = value


"""
Counter
    add
    createCounter
"""
def create(port):
    request = COUNTER.CreateCounterRequest()
    msg = port.createCounter(request)
    print "----- create ----- "
    print "\tAddress -- ", msg._EndpointReference._Address
    print "\tRefProp -- ", msg._EndpointReference._ReferenceProperties
    print "\t\tany -- ", msg._EndpointReference._ReferenceProperties._any
    return msg

def add(port, value=999):
    print "----- add ----- "
    request = COUNTER.AddInputMessage(value)
    msg = port.add(request)
    print "RESPONSE: ", msg

"""
ResourceLifetime
    Destroy
    GetCurrentMessage
    SetTerminationTime
"""
def Destroy(port):
    print "----- destroy ----- "
    request = COUNTER.DestroyRequest()
    msg = port.Destroy(request)
    print "RESPONSE: ", msg

def SetTerminationTime(port, interval):
    print "----- SetTerminationTime ----- "
    request = COUNTER.SetTerminationTimeRequest()
    request._RequestedTerminationTime = time.gmtime(time.time()+interval)
    msg = port.SetTerminationTime(request)
    print "TerminationTime: ", msg._NewTerminationTime
    print "CurrrentTime:    ", msg._CurrentTime


"""
ResourceProperties
    GetResourceProperty
"""
def GetResourceProperty(port):
    print "----- GetResourceProperty ----- "
    pyclass = COUNTER_SERVICE.GetResourcePropertyRequest
    request = pyclass(("http://counter.com", "Value"))
    msg = port.GetResourceProperty(request)
    print "RESPONSE: ", msg._any


def Subscribe(port, consumer):
    """
    port -- Subscription Client
    consumer -- Notification Consumer
    Need to specify the prefix-namespace mapping in the
    XML instance to deal with QNames passed via the 
    TopicExpression mixed text content.
    """
    print "----- Subscribe ----- "
    
    request = COUNTER.SubscribeRequest()
    epr = consumer.context.getConsumerReference()
    request._ConsumerReference = epr
    request._TopicExpression = TopicExpression()
    request._TopicExpression.setMixed("counter:Value")
    request._UseNotify = True
    port.binding.nsdict['counter'] = "http://counter.com"
    msg = port.Subscribe(request)
    print "RESPONSE: ", msg


def main():
    netloc = '%s:%d'%(OPTIONS.host,OPTIONS.port)
    path = CP.get(SEC_PATHS, 'CounterService')
    kw = {}
    if OPTIONS.debug:
        kw['tracefile'] = sys.stdout
        log.startLogging(sys.stdout)

    scheme = 'http'
    if OPTIONS.ssl:
        scheme = 'https'
        from pyGlobus.io import GSIHTTPConnection
        from pyGlobus.ioc import GLOBUS_IO_SECURE_CHANNEL_MODE_SSL_WRAP,\
            GLOBUS_IO_SECURE_DELEGATION_MODE_NONE
        kw['transport'] = GSIHTTPConnection
        kw['transdict'] = \
            {'channel_mode':GLOBUS_IO_SECURE_CHANNEL_MODE_SSL_WRAP,
             'delegation_mode':GLOBUS_IO_SECURE_DELEGATION_MODE_NONE,
            }

    urlTuple = (scheme,netloc,path,'','','')
    url = urlparse.urlunparse(urlTuple)
    locator = COUNTER.CounterServiceLocator()
    port = locator.getCounterPortType(portAddress=url, **kw)

    response = create(port)

    # Get New port w/ EPR
    port = locator.getCounterPortType(portAddress=url, \
        endPointReference=response._EndpointReference, \
        **kw)

    consumer = NotificationConsumer()
    consumer.start()

    print "=======> Queue Events.."
    reactor.callWhenRunning(Subscribe, port, consumer)
    reactor.callWhenRunning(add, port, 13)
    reactor.callLater(5, reactor.stop)

 
if __name__ == '__main__':
    reactor.callWhenRunning(main)
    reactor.run()



