############################################################################
# Joshua R. Boverhof, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################
import os, socket
from ConfigParser import ConfigParser
from twisted.application import service, internet
from twisted.protocols import wire
from twisted.internet import protocol
from twisted.web.server import Site
from twisted.web.resource import Resource
from twisted.python.log import startLogging

from pyGridWare.generated.types import *
from pyGridWare.services.CounterService import Counter
from pyGridWare.services.SecureCounterService import SecureCounter
from pyGridWare.services.SecureConversationService import SecureConversationService
from pyGridWare.services.SubscriptionManager import SubscriptionManager

from pyGridWare.utility.http import ChunkedHTTPChannel


###########################################################################
# Grab Port from configuration file
#
###########################################################################
CP = ConfigParser()
CP.read('config.txt')
PORT = CP.getint('Location', 'port')
HOST = socket.gethostbyname(socket.gethostname())
DEBUG = CP.getint('logging', 'debug')
SSL = CP.getint('security', 'ssl')
CERTFILE = CP.get('security', 'certfile') or '%s/%s' %(os.environ['HOME'],'.globus/usercert.pem')
KEYFILE = CP.get('security', 'keyfile') or '%s/%s' %(os.environ['HOME'],'.globus/userkey.pem')

###########################################################################
# Debugging
#
###########################################################################
if DEBUG:
    from ZSI.wstools.logging import setBasicLoggerDEBUG
    setBasicLoggerDEBUG()
    f = open(CP.get('logging', 'serverlog'), 'w+')
    startLogging(f)

###########################################################################
# Get Resources
#
###########################################################################
def GetResource():
    root = Resource()
    root.putChild('wsrf', Resource())

    resource = Resource()
    root.getStaticEntity('wsrf').putChild('services', resource)

    resource.putChild('CounterService', 
        Counter(post='/wsrf/services/CounterService'))

    resource.putChild('SecureCounterService-authService', 
        SecureConversationService(post='/wsrf/services/SecureCounterService-authService'))

    resource.putChild('SecureCounterService', 
        SecureCounter(post='/wsrf/services/SecureCounterService'))

    # Need to Set URL for SubscriptionManager, need to know
    # its URL before it has received a request.
    sub_path = '/wsrf/services/SubscriptionManagerService'
    subResource = SubscriptionManager.getInstance(post=sub_path)
    subResource.setServiceURL('http://%s:%d/wsrf/services/SubscriptionManagerService' %(HOST,PORT))
    if SSL:
        subResource.setServiceURL('https://%s:%d/wsrf/services/SubscriptionManagerService' %(HOST,PORT))
    resource.putChild('SubscriptionManagerService', subResource)
    return root


###########################################################################
# Set up Application
#
###########################################################################

# swap in the chunked request aware http channel handler
Site.protocol = ChunkedHTTPChannel
myFactory = Site(GetResource())

port = None
if SSL:
    from twisted.internet import ssl
    port = internet.SSLServer(PORT, myFactory, 
              ssl.DefaultOpenSSLContextFactory(KEYFILE, CERTFILE))
else:
    port = internet.TCPServer(PORT, myFactory)

application = service.Application("WSRF-Container")
port.setServiceParent(application)


