############################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
import ZSI.TC
from twisted.python import log
from pyGridWare.properties.ResourceProperties import ResourceProperties
from pyGridWare.generated.types import GetNSContainer
import pyGridWare.generated.types.Counter_services_types

_TARGETNAMESPACE = "http://counter.com"
counter_com = GetNSContainer(pyGridWare.generated.types.Counter_services_types,
                  _TARGETNAMESPACE)

class CounterKey(int):
    typecode = ZSI.TC.Iint(pname=(_TARGETNAMESPACE,'CounterKey'))


class CounterProperties(ResourceProperties):
    '''Associate a TopicSpace with every targetNamespace, generate a notification messsage
    when observe a resource property change.  Associate the notification message with a
    topic, and push to NotificationProducer.  NP will look for the topic in its Topic Set,
    if find a Subscription matching send a "Notify" message.

    ??
    <wstop:topicSpace name="ResourcePropertiesTopicSpace" targetNamespace="http://counter.com">
        <wstop:topic name="CounterResourceProperty">
            <wstop:topic name="Value" messageTypes="SomeNotificationMessage"/>
        </wstop:topic>
    </wstop:topicSpace>

    or

    <wstop:topicSpace name="ResourcePropertiesTopicSpace" targetNamespace="http://counter.com">
        <wstop:topic name="Value" messageTypes="SomeNotificationMessage"/>
        <wstop:topic name="terminationTime" messageTypes="SomeNotificationMessage"/>
        <wstop:topic name= bla bla.../>
    </wstop:topicSpace>
    '''
    declaration = counter_com.CounterRP_Dec

    def getValue(self):
        '''convience method for accessing Value
        '''
        return self.getResourceProperty(_TARGETNAMESPACE, 'Value')

    def setValue(self, value):
        '''convience method for setting Value
        '''
        self.setResourceProperty(_TARGETNAMESPACE, 'Value', value)


