############################################################################
# Joshua R. Boverhof, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################
from twisted.python import log
from twisted.internet import reactor
from twisted.web.server import Site
from twisted.web.resource import Resource
from twisted.internet.error import CannotListenError

from pyGridWare.generated.resource.NotificationConsumer_context import NotificationConsumerContext
from pyGridWare.generated.services.WS_BaseNotification_services_server import WS_BaseNotification as NCServiceInterface
import pyGridWare.generated.types.NotificationConsumer_services_types

# Enable HTTP chunking
from pyGridWare.utility.http import ChunkedHTTPChannel
Site.protocol = ChunkedHTTPChannel

class NotificationConsumerService(NCServiceInterface):

    def __init__(self, post='/wsrf/services/NotificationConsumerService', **kw):
        NCServiceInterface.__init__(self, post)

    def wsa_Notify(self, ps, address):
        response = NCServiceInterface.wsa_Notify(self, ps, address)

        log.msg("Notification: ", self.request)
        pyobj = self.request._NotificationMessage[0]
        log.msg("  NotificationMessage: ", pyobj)
        log.msg("    Topic: ", pyobj._Topic)
        log.msg("    ProducerReference: ", pyobj._ProducerReference)
        log.msg("    Message: ", pyobj._Message)
        pyobj = pyobj._Message._ResourcePropertyValueChangeNotification
        log.msg("        NewValue: ", pyobj._NewValue)
        log.msg("          Value: ", pyobj._NewValue._any)
        if pyobj._OldValue is not None:
            log.msg("        OldValue: ", pyobj._OldValue)
            log.msg("          Value: ", pyobj._OldValue._any)
        return response


class NotificationConsumerTwisted:

    def __init__(self, path='/wsrf/services/NotificationConsumerService', 
        notificationClass=NotificationConsumerService):
        '''
        notificationClass -- Notification Consumer Service 
        '''
        self.context = NotificationConsumerContext()
        self.path = path
        self.notication_consumer = None
        self.factory = None
        self.port = None

        assert issubclass(notificationClass, NCServiceInterface), \
            'notificationClass must be NotificationConsumer'
        root = Resource()
        resource = root
        token = '/'
        ipath = path.strip(token)
        for key in ipath.split(token)[:-1]:
            resource.putChild(key, Resource())
            resource = resource.getStaticEntity(key)

        service_key = ipath.split(token)[-1]
        self.notication_consumer = notificationClass(post=path)
        resource.putChild(service_key, self.notication_consumer)
        self.factory = Site(root)

    def getNotificationContext(self):
        return self._ctx
    def setNotificationContext(self, context):
        self._ctx = context
    context = property(getNotificationContext, setNotificationContext, \
       None, 'NotificationConsumer Context')

    def start(self):
        try:
            #internet.TCPServer(self.port, self.factory).setServiceParent(self.application)
            self.port = reactor.listenTCP(0, self.factory)
        except CannotListenError, ex:
            raise
        log.debug("NotificationConsumer: ", self.port.getHost())
        url = 'http://%s:%d%s' %(self.port.getHost().host,self.port.getHost().port,self.path)
        self.notication_consumer.setServiceURL(url)
        self.context.setAddress(address=url)

NotificationConsumer = NotificationConsumerTwisted


