############################################################################
# Joshua R. Boverhof, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################
import sys
from twisted.python import log
from pyGridWare.patterns import Singleton
from pyGridWare.addressing.AddressingUtils import AddressingUtils
from pyGridWare.notification.NotificationProducer import NotificationProducer
from pyGridWare.resource.ResourceContext import ResourceContext
from pyGridWare.lifetime.ResourceLifetime import ImmediateResourceTermination,\
    ScheduledResourceTermination

from pyGridWare.generated.types import GetNSContainer
import pyGridWare.generated.types.WS_BaseNotification_services_types
from pyGridWare.generated.stubs.WS_BaseNotification_services import \
    WS_BaseNotificationServiceLocator, NotificationConsumerSOAPBindingSOAP

from pyGridWare.generated.resource.SubscriptionProducer_context import \
    SubscriptionContext

_BaseNotification = \
    GetNSContainer(pyGridWare.generated.types.WS_BaseNotification_services_types,
        "http://docs.oasis-open.org/wsn/2004/06/wsn-WS-BaseNotification-1.2-draft-01.xsd")

class Topic:
    '''A Topic must specify a "Dialect", which is used to
    interpret the "expression" passed via mixed or <any>.
    '''
    simple = "http://docs.oasis-open.org/wsn/2004/06/TopicExpression/Simple"
    def __init__(self, topic):
        '''
        element wildcard
        mixed content
        attributes: Dialect
        '''
        self.topic = topic

    def getDialect(self):
        return self.topic._attrs['Dialect']

    def getRootTopic(self):
        return self.topic._text

    def evaluate(self, event):
        '''XXX Need DOM to resolve prefix
        The only valid token for 'Simple' dialect is a qName, but
        sadly this is not represented as a qName in XMLSchema.  Instead
        it is only understood to be a qName in this layer.
        '''
        if self.simple != self.getDialect():
            raise NotImplementedError, 'Dialect %s currently not supported' %self.dialect
        qualifiedName = self.getRootTopic()
   

class Subscription(object):
    '''Represents a subscription, registered to NotificationProducer
    '''
    def __init__(self, context, producer):
        '''
        context -- SubscriptionContext instance.
        producer -- Monitoring Resource Context 
        '''
        self.context = context
        self.producer = producer
        self.request = None
        self.topic = None

    def getContext(self):
        return self.__sub_context
    def setContext(self, context):
        assert isinstance(context, SubscriptionContext), \
            'resource must be a SubscriptionContext instance'
        self.__sub_context = context
    context = property(getContext, setContext, None,
        'Subscription Context')
 
    def getProducerContext(self):
        return self.__p_context
    def setProducerContext(self, context):
        assert isinstance(context, ResourceContext), \
            'resource must be a ResourceContext instance'
        self.__p_context = context
    producer = property(getProducerContext, setProducerContext, None,
        'Subscribed to Resource Context')

    def getSubscriptionRequest(self):
        return self.__request
    def setSubscriptionRequest(self, request):
        #assert request is None or isinstance(request, _BaseNotification.Subscribe_Dec), \
        #    'request(%s) must be a Subscribe declaration (%s,%s)' \
        #    %(request, _BaseNotification.Subscribe_Dec.schema,_BaseNotification.Subscribe_Dec.literal)
        self.__request = request
    request = property(getSubscriptionRequest, setSubscriptionRequest, None,
        'Subscription Request')

    def getTopicExpression(self):
        return self.__topic
    def setTopicExpression(self, topic):
        assert topic is None or (hasattr(topic, '_attrs') and topic._attrs.has_key('Dialect')), \
            'topic must have Dialect attribute'
        self.__topic = topic
    topic = property(getTopicExpression, setTopicExpression, None,
        'TopicExpression')

    def getConsumerReference(self):
        return self.__reference
    def setConsumerReference(self, reference):
        self.__reference = reference
    reference = property(getConsumerReference, setConsumerReference, None,
            'EPR (%s,%s)' %(_BaseNotification.ConsumerReference_Dec.schema,_BaseNotification.ConsumerReference_Dec.literal))

    def setUp(self, request):
        '''
        request -- ConsumerReference element declaration, contents
            ConsumerReference, TopicExpression, UseNotify
        '''
        self.request = request
        if self.request._UseNotify is True:
            self.reference = self.request._ConsumerReference
            url = self.reference._Address
            self.topic = self.request._TopicExpression
            self.notificationProducer = NotificationProducer()
            self.notificationProducer.setUp(self)

