############################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
from ZSI import SplitQName
from twisted.python import log

class Topic:

    def __call__(self, *args, **kw):
        return


class ResourceProperties:
    '''
    ResourceProperties must be a GED, and it contains any
    number references to other GEDs.
    '''
    declaration = None

    def __init__(self):
        self.topicListeners = []
        self._element = self.GetDeclarationClass()()
   
    #def addTopicListener(self, method, *args):
    def addTopicListener(self, method, topic):
        ''' Need to hold onto DOM to make this really work
        specification is retarded, uses text content to pass
        in a qualifiedName.
        '''
        assert callable(method), 'Callback must be callable.'
        if not getattr(topic, '_text', None): return
        prefix,name = SplitQName(topic._text)
        log.debug("TOPIC: %s:%s" %(prefix,name))

        def cb(value):
            if value.typecode.pname != name: return
            return method(value)

        self.topicListeners.append(cb)

    def getResourceProperty(self, namespaceURI, localName):
        '''Resource must be a global element declaration, 
        and so assume that it is either the ResourceProperty
        element itself, or one of the element references in
        its model group.
        '''
        what = self._element
        if what.nspname == namespaceURI and  what.pname == localName:
            log.debug("TOP-WHAT(%s, %s): " %(what.nspname,what.pname))
            return what
        elif hasattr(what, 'ofwhat'):
            for what in what.ofwhat:
                log.debug("WHAT(%s, %s): " %(what.nspname,what.pname))
                if what.nspname == namespaceURI and  what.pname == localName:
                    element = getattr(self._element, what.aname, None)
                    log.debug("ELEMENT: ", element)
                    if element is not None:
                        class Wrapper(type(element)):
                            typecode = what
                        return Wrapper(element)
                    return None
        raise NoResourceProperty, \
            "getResourceProperty failed to locate (%s,%s)" %(namespaceURI,localName)

    def setResourceProperty(self, namespaceURI, localName, pyobj):
        what = self._element
        for what in what.ofwhat:
            if what.nspname == namespaceURI and  what.pname == localName:
                break
        else:
            raise RuntimeError, 'Can not setResourceProperty (%s,%s)' %(what.nspname,what.pname)

        setattr(self._element, what.aname, pyobj)
        log.debug('--> TopicListeners: %s' %self.topicListeners)

        class Wrapper(type(pyobj)):
            typecode = what

        for cb in self.topicListeners:
            log.debug('--> Callback')
            cb(Wrapper(pyobj))

    def GetDeclarationClass(cls):
        return cls.declaration
    GetDeclarationClass = classmethod(GetDeclarationClass)


"""
  <wsdl:portType name="GetResourceProperty">
    <wsdl:operation name="GetResourceProperty">
      <wsdl:input  name="GetResourcePropertyRequest" 
                   message="wsrp:GetResourcePropertyRequest" />
      <wsdl:output name="GetResourcePropertyResponse" 
                   message="wsrp:GetResourcePropertyResponse" />

  <wsdl:portType name="GetMultipleResourceProperties">
    <wsdl:operation name="GetMultipleResourceProperties">
      <wsdl:input  name="GetMultipleResourcePropertiesRequest" 
                   message="wsrp:GetMultipleResourcePropertiesRequest" />
      <wsdl:output name="GetMultipleResourcePropertiesResponse" 
                   message="wsrp:GetMultipleResourcePropertiesResponse" />

  <wsdl:portType name="SetResourceProperties">
    <wsdl:operation name="SetResourceProperties">
      <wsdl:input  name="SetResourcePropertiesRequest" 
                   message="wsrp:SetResourcePropertiesRequest" />
      <wsdl:output name="SetResourcePropertiesResponse" 
                   message="wsrp:SetResourcePropertiesResponse" />

  <wsdl:portType name="QueryResourceProperties">
    <wsdl:operation name="QueryResourceProperties">
      <wsdl:input  name="QueryResourcePropertiesRequest" 
                   message="wsrp:QueryResourcePropertiesRequest" />
      <wsdl:output name="QueryResourcePropertiesResponse" 
                   message="wsrp:QueryResourcePropertiesResponse" />
"""

class ResourcePropertiesInterface:

    def GetResourceProperty(resource, namespaceURI, localName):
        ''' <xsd:element name="GetResourceProperty" type="xsd:QName" />
        '''
        resource.getResourceProperty(namespaceURI, localName)
    GetResourceProperty = staticmethod(GetResourceProperty)

    def GetMultipleResourceProperties(resource):
        pass
    GetMultipleResourceProperties = staticmethod(GetMultipleResourceProperties)

    def SetResourceProperties(resource):
        pass
    SetResourceProperties = staticmethod(SetResourceProperties)

    def QueryResourceProperties(resource):
        pass
    QueryResourceProperties = staticmethod(QueryResourceProperties)



