############################################################################
# Joshua R. Boverhof, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################
import time
from ZSI.wstools.Namespaces import OASIS

class ResourceContext(object):
    __slots__ = ['_url','_terminationTime','_resourceID','_properties']
    home = None
    
    def __init__(self):
        self._resourceID = None
        self._properties = None
        self._terminationTime = None
        self.url = None

    def getAddress(self):
        return self._url
    def setAddress(self, address):
        self._url = address
    url = property(getAddress, setAddress, None,
       'URL of service')

    def getResourceID(self):
        return self._resourceID
    def setResourceID(self, resourceID):
        self._resourceID = resourceID
    #resourceID = property(getResourceID, setResourceID, None, 
    #   'Resource Identifier')

    def getResourceProperties(self):
        return self._properties
    def setResourceProperties(self, properties):
        self._properties = properties
    properties = property(getResourceProperties, setResourceProperties, None,
       'Resource Properties')

    def getTerminationTime(self):
        '''return a UTC float
        '''
        return self._terminationTime
    def setTerminationTime(self, terminationTime):
        '''Do not set _terminationTime!!!
        store as a UTC float, and in properties as a gmtime tuple.
        '''
        assert terminationTime is None or isinstance(terminationTime, float), \
            'ResourceContext, terminationTime expecting a float or None'
        self._terminationTime = terminationTime
        terminationTime = tuple(time.gmtime(terminationTime))
        self.properties.setResourceProperty(OASIS.LIFETIME, 'TerminationTime', terminationTime)

    # class methods
    def getResourceHome(cls):
        return cls.home
    getResourceHome = classmethod(getResourceHome)



