############################################################################
# Matt Rodriguez, LBNL
# See Copyright for copyright notice!
###########################################################################
from ZSI.wstools.Namespaces import SOAP, OASIS, DSIG
import sha

class SignatureHandlerError(RuntimeError):
    pass

class BaseSignatureHandler:
    """
    The parent class for all of the Signature handlers.
    This class will handle tasks which are not specific
    to gss signing
    """
    tokenRef = 10101
    def signHeaderElements(self, sw):
        processorNss={'SOAP-ENV':SOAP.ENV}
        sw.dom.setContext(processorNss=processorNss)
        nodes = sw.dom.evaluate(expression='//SOAP-ENV:Header')
        headerNode = nodes[0]
        headerNodes = headerNode.getElements(None, None)
        for node in headerNodes:
            idVal = self.getTokenRefString()
            node.setAttributeNS(OASIS.UTILITY, 'Id', idVal)


    def getDigestValueFromSoapWriter(self, sw, uri):
        '''
<ns1:createCounter xmlns:SOAP-ENC="http://schemas.xmlsoap.org/soap/encoding/" xmlns:SOAP-ENV="http://schemas.xmlsoap.org/soap/envelope/" xmlns:ZSI="http://www.zolera.com/schemas/ZSI/" xmlns:ns1="http://counter.com" xmlns:ns2="http://schemas.xmlsoap.org/ws/2002/04/utility" xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" ns2:Id="digestSource"></ns1:createCounter>
        '''
        digExpString = '//*[@wsu:Id="' + uri[1:] +'"]'
        if not sw.dom.isContextInitialized():
            raise SignatureHandlerError('SoapWriter ElementProxy has no context')
        nodes = sw.dom.evaluate(expression=digExpString)
        node = nodes[0]
        cxml = node.canonicalize()
        digestValue = sha.sha(cxml).digest()
        return digestValue

    def getDigestValueFromDom(self, element, uri, algorithm=DSIG.C14N):
        uriExpString = '//*[@wsu:Id="' + uri[1:] +'"]'
        if not element.isContextInitialized():
            raise SignatureHandlerError('ElementProxy has no context')
        uriNodes = element.evaluate(expression=uriExpString)
        if len(uriNodes) != 1:
            raise VerifyError("more than one uri found with uri: " + uri)
        uriNode = uriNodes[0]
        cxml = uriNode.canonicalize(algorithm)
        digestValue = sha.sha(cxml).digest()
        return digestValue

    def getTokenRefString(self):
        """
        Return an unique number to identify an element
        for the element's id attribute
        """
        BaseSignatureHandler.tokenRef +=1
        return str(BaseSignatureHandler.tokenRef)
