############################################################################
# Joshua R. Boverhof, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################
import time
from pyGridWare.addressing.AddressingUtils import \
    AddressingUtils
from pyGridWare.lifetime.ResourceLifetime import \
    ImmediateResourceTermination, ScheduledResourceTermination
from pyGridWare.services.SubscriptionManager import \
    SubscriptionManager
from pyGridWare.generated.resource.Counter_context import \
    CounterContext
from pyGridWare.generated.services.Counter_services_server import \
    Counter as CounterBindings
from pyGridWare.generated.resource.Counter_context import \
    CounterContext


class Counter(CounterBindings):

    def __init__(self, post='/wsrf/services/CounterService', **kw):
        CounterBindings.__init__(self, post)

    def wsa_Destroy(self, ps, address):
        # Get request and response
        response = CounterBindings.wsa_Destroy(self, ps, address)
        self.request

        # Get Resource Context
        home = CounterContext.getResourceHome()
        ctx = home.getResourceContext(ps, address)

        # Terminate Resource
        ImmediateResourceTermination.Destroy(ctx)

        return response

    def wsa_GetCurrentMessage(self, ps, address):
        # Get request and response
        response = CounterBindings.wsa_GetCurrentMessage(self, ps, address)
        self.request

        # Get Resource Context
        home = CounterContext.getResourceHome()
        ctx = home.getResourceContext(ps, address)

        return response

    def wsa_GetResourceProperty(self, ps, address):
        # Get request and response
        response = CounterBindings.wsa_GetResourceProperty(self, ps, address)
        self.request

        # Get Resource Context
        home = CounterContext.getResourceHome()
        ctx = home.getResourceContext(ps, address)

        properties = ctx.properties
        namespaceURI,element = self.request
        property = properties.getResourceProperty(namespaceURI,element)

        # assign return values to response object
        response._any = property
        return response

    def wsa_SetTerminationTime(self, ps, address):
        # Get request and response
        response = CounterBindings.wsa_SetTerminationTime(self, ps, address)
        self.request

        # Get Resource Context
        home = CounterContext.getResourceHome()
        ctx = home.getResourceContext(ps, address)

        # Schedule Termination
        ScheduledResourceTermination.SetTerminationTime(ctx, self.request._RequestedTerminationTime)

        # assign return values to response object
        response._NewTerminationTime = time.gmtime(ctx.getTerminationTime())
        response._CurrentTime = time.gmtime(time.time())
        return response

    def wsa_Subscribe(self, ps, address):
        # Get request and response
        response = CounterBindings.wsa_Subscribe(self, ps, address)
        self.request

        # Get Resource Context
        home = CounterContext.getResourceHome()
        ctx = home.getResourceContext(ps, address)

        # Pass SubscriptionRequest off
        assert SubscriptionManager._isInstantiated(), 'SubscriptionManager is not instantiated'
        manager = SubscriptionManager.getInstance()
        response._SubscriptionReference = manager.subscribe(self.request, ctx)
        return response

    def wsa_add(self, ps, address):
        # Get request and response
        response = CounterBindings.wsa_add(self, ps, address)
        self.request

        # Get Resource Context
        home = CounterContext.getResourceHome()
        ctx = home.getResourceContext(ps, address)

        # Access resource properties, and evaluate new value.
        counter = ctx.properties
        result = counter.getValue()
        result += self.request
        counter.setValue(result)

        # response is immutable instance, create new one with result.
        response = response.__class__(result)
        return response

    def wsa_createCounter(self, ps, address):
        # Get request and response
        response = CounterBindings.wsa_createCounter(self, ps, address)
        self.request

        # Create Resource Context
        ctx = CounterContext()
        url = self.getServiceURL()
        ctx.setAddress(url)
        home = ctx.getResourceHome()
        key = home.create()
        home.add(ctx, key)
        ctx.setResourceID(key)

        # Initialize Value to 0
        counter = ctx.properties
        counter.setValue(0)

        # Create EndPointReference and set in response
        epr = AddressingUtils.createEndpointReference(ctx)
        response._EndpointReference = epr
        return response


