############################################################################
# Joshua R. Boverhof, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################
import time

from twisted.internet import reactor

from pyGridWare.generated.services.GlobusRunService_services_server import *
from pyGridWare.generated.resource.GlobusRun_context import *

from pyGridWare.addressing.AddressingUtils import AddressingUtils
from pyGridWare.lifetime.ResourceLifetime import \
    ImmediateResourceTermination, ScheduledResourceTermination

from pyGridWare.services.SubscriptionManager import \
    SubscriptionManager

from twisted.internet.protocol import ProcessProtocol
from twisted.internet.error import ProcessDone, ProcessTerminated
from twisted.python.failure import Failure

class GlobusRunProcessProtocol(ProcessProtocol):
    def __init__(self, rp):
        self.rp = rp

    def connectionMade(self):
        print "connectionMade"
        self.rp.set_status("connectioNMade")
        self.transport.closeStdin()
        
    def inConnectionLost(self):
        print "inConnectionLost"
        self.rp.set_status("inConnectionLost")

    def outConnectionLost(self):
        print "outConnectionLost"
        self.rp.set_status("outConnectionLost")
        
    def outReceived(self, data):
        print "outReceived"
        self.rp.set_status("outReceived, data: %s" % data)
        
    def errReceived(self, data):
        print "errReceived"
        self.rp.set_status("errReceived, data: %s" % data)

    def processEnded(self, status):
        print "processEnded", status.value.exitCode
        
        if isinstance(status.value, ProcessDone):
            self.rp.set_status("processEnded (terminated normally)")

        elif isinstance(status.value, ProcessTerminated):
            self.rp.set_status("processEnded (terminated abnormally: status: %d)" % \
                               status.value.exitCode)
            
        print "EXIT CODE: ", type(status.value.exitCode)
        print "RP: ", self.rp
        self.rp.set_returnCode(0)
        
class GlobusRun(GlobusRunService):

    def __init__(self, post='/services/GlobusRunService', **kw):
        GlobusRunService.__init__(self, post, **kw)

    def wsa_Subscribe(self, ps, address):
        # Get request and response
        response = GlobusRunService.wsa_Subscribe(self, ps, address)

        # Get Resource Context
        home = GlobusRunContext.getResourceHome()
        ctx = home.getResourceContext(ps, address)

        # Pass SubscriptionRequest off
        assert SubscriptionManager._isInstantiated(), 'SubscriptionManager is not instantiated'
        manager = SubscriptionManager.getInstance()
        response._SubscriptionReference = manager.subscribe(self.request, ctx)
        return response

    def wsa_create(self, ps, address):
        response = GlobusRunService.wsa_create(self, ps, address)

        # Create Resource Context
        ctx = GlobusRunContext()
        url = self.getServiceURL()
        ctx.setAddress(url)
        home = ctx.getResourceHome()
        key = home.create()
        home.add(ctx, key)
        ctx.setResourceID(key)

        # Access resource properties, and set status 0
        rp = ctx.properties
        rp.set_status('INITIAL')
        rp.set_returnCode(7)
        args = [self.request._executable]
        args.extend(self.request._argument)
        reactor.spawnProcess(GlobusRunProcessProtocol(rp), self.request._executable, self.request._argument)

        # Create EndPointReference and set in response
        epr = AddressingUtils.createEndpointReference(ctx)
        response._EndpointReference = epr
        return response

    def wsa_Destroy(self, ps, address):
        response = GlobusRunService.wsa_Destroy(self, ps, address)

        # Get Resource Context
        home = GlobusRunContext.getResourceHome()
        ctx = home.getResourceContext(ps, address)

        # Access resource properties, and set status 0
        #print "GLOBUSRUN --- DESTROY"
        #rp = ctx.properties
        #rp.set_status('DESTROY')
        #rp.set_returnCode(666)

        # Terminate Resource
        #ImmediateResourceTermination.Destroy(ctx)
        return response

    def wsa_getMultipleResourceProperties(self, ps, address):
        response = GlobusRunService.wsa_getMultipleResourceProperties(self, ps, address)

        # Get Resource Context
        home = GlobusRunContext.getResourceHome()
        ctx = home.getResourceContext(ps, address)

        namespaceURI,name = self.request
        property = ctx.properties.getResourceProperty(namespaceURI,name)

        # assign return values to response object
        response._any = property
        return response

    def wsa_GetResourceProperty(self, ps, address):
        response = GlobusRunService.wsa_GetResourceProperty(self, ps, address)

        # Get Resource Context
        home = GlobusRunContext.getResourceHome()
        ctx = home.getResourceContext(ps, address)

        print "-"*90
        print "GET REQUEST: ", self.request
        print "-"*90
        namespaceURI,name = self.request
        property = ctx.properties.getResourceProperty(namespaceURI,name)

        # assign return values to response object
        response._any = property
        return response

    def wsa_queryResourceProperties(self, ps, address):
        response = GlobusRunService.wsa_queryResourceProperties(self, ps, address)
        return response

    def wsa_setResourceProperty(self, ps, address):
        response = GlobusRunService.wsa_setResourceProperty(self, ps, address)
        return response

    def wsa_setTermTime(self, ps, address):
        response = GlobusRunService.wsa_setTermTime(self, ps, address)
        self.request

        # Get Resource Context
        home = GlobusRunContext.getResourceHome()
        ctx = home.getResourceContext(ps, address)

        # Schedule Termination
        ScheduledResourceTermination.SetTerminationTime(ctx, self.request._RequestedTerminationTime)

        # assign return values to response object
        response._NewTerminationTime = time.gmtime(ctx.getTerminationTime())
        response._CurrentTime = time.gmtime(time.time())
        return response



