############################################################################
# Joshua R. Boverhof, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################
import sys
from twisted.python import log
from pyGridWare.patterns import Singleton
from pyGridWare.addressing.AddressingUtils import AddressingUtils
from pyGridWare.resource.ResourceContext import ResourceContext
from pyGridWare.lifetime.ResourceLifetime import ImmediateResourceTermination,\
    ScheduledResourceTermination
from pyGridWare.notification.SubscriptionManager import Topic, Subscription

from pyGridWare.generated.types import GetNSContainer
import pyGridWare.generated.types.WS_BaseNotification_services_types

from pyGridWare.generated.resource.SubscriptionProducer_context import \
    SubscriptionContext
from pyGridWare.generated.services.SubscriptionManagerService_services_server import \
    SubscriptionManagerService

ns6 = GetNSContainer(pyGridWare.generated.types.WS_BaseNotification_services_types,
          "http://docs.oasis-open.org/wsn/2004/06/wsn-WS-BaseNotification-1.2-draft-01.xsd")

class SubscriptionManager(SubscriptionManagerService, Singleton):
    '''
    service -- SubscriptionManagerService.
    subscriptions -- list of Subscriptions.
    '''
    service = None
    subscriptions = []

    def __init__(self, post='/wsrf/services/SubscriptionManagerService', **kw):
        SubscriptionManagerService.__init__(self, post, **kw)

    def NewSubscription(cls, request, ctx, subCtx):
        '''Create a new subscription
        request -- subscription request, containing ConsumerReference and TopicExpression.
        ctx -- resource context subscription applies to.
        subCtx -- subscription resource context.
        '''
        sub = Subscription(subCtx, ctx)
        sub.setUp(request)
        cls.subscriptions.append(sub)
    NewSubscription = classmethod(NewSubscription)
 
    def Destroy(ps, address):
        home = SubscriptionContext.getResourceHome()
        ctx = home.getResourceContext(ps, address)
        ImmediateResourceTermination.Destroy(ctx)
    Destroy = staticmethod(Destroy)

    def subscribe(self, request, ctx):
        '''Returns an EPR representing a SubscriptionReference.
        request -- subscription request, containing ConsumerReference and TopicExpression.
        ctx -- resource context subscription applies to.
        '''
        assert isinstance(ctx, ResourceContext), \
            'resource must be a ResourceContext instance'

        # Subscription Context
        sub_ctx = SubscriptionContext()
        sub_ctx.setAddress(self.getServiceURL())
        sub_home = sub_ctx.getResourceHome()
        sub_key = sub_home.create()
        sub_home.add(sub_ctx, sub_key)
        sub_ctx.setResourceID(sub_key)

        self.NewSubscription(request, ctx, sub_ctx)
        epr = AddressingUtils.createEndpointReference(sub_ctx)
        return epr

    # Callback methods
    def wsa_Destroy(self, ps, address):
        response = SubscriptionManagerService.wsa_Destroy(self, ps, address)
        self.Destroy(ps, address)
        # assign return values to response object
        return response


