############################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
import os, sys, string, urlparse
from distutils.core import setup
from distutils.spawn import spawn
from distutils.dist import Distribution
from os.path import isfile
from ConfigParser import ConfigParser
import ZSI
import ZSI.wsdlInterface
from ZSI.wstools.WSDLTools import WSDLReader
from ZSI.generate.wsdl2python import WriteServiceModule
from ZSI.generate.wsdl2python import Config


GLOBUS_LOCATION = os.environ.get('GLOBUS_LOCATION')
ALLOW_SLOPPY_WSDL = True
CONFIG = 'config.txt'
GLOBUS_SECTION = 'GLOBUS_WSDL'
LOCAL_SECTION = 'LOCAL_WSDL'
DEBUG = False

def WriteServiceFile(wsdl):
    ss = ServerCallbackDescription()
    ss.fromWsdl(wsdl)
    file_name =  'pyGridWare/generated/services/%s.py' %ss.getStubName()
    fd = open(file_name, 'w+')
    ss.write(fd)
    fd.close()

def BetaWriteStubsTypes(wsdl):
    wsm = WriteServiceModule(wsdl, addressing=True)
    wsm.set_module_path('pyGridWare/generated/types')
    wsm.write(True)

def BetaWriteStubs(wsdl):
    domletteLib = \
    'from pyGridWare.utility.dom import DomletteReader, DomletteElementProxy'

    Config.readerclass = 'DomletteReader'
    Config.writerclass = 'DomletteElementProxy'
    Config.serviceImports.append(domletteLib)

    wsm = WriteServiceModule(wsdl, addressing=True)
    wsm.set_module_path('pyGridWare/generated/stubs')
    wsm.gatherNamespaces()
    
    f_types, f_services = wsm.get_module_names()
    
    f_types = f_types.replace('/','.')
    f_types = f_types.replace('.stubs.', '.types.')
    
    fd = open(f_services + '.py', 'w+')
    wsm.write_services(f_types, f_services, fd, True)
    fd.close()

    Config.clear()
    

def GetWSDL(url, isGlobus=True):
    file = url
    if isGlobus:
        assert GLOBUS_LOCATION is not None, 'Must set GLOBUS_LOCATION to regenerate globus modules'
        location = GLOBUS_LOCATION + '/'
        file = urlparse.urljoin(location, url)
    reader = WSDLReader()
    if isfile(file) is False:
        return reader.loadFromURL(file)
    return reader.loadFromFile(file)


def BetaRegenerateModules(services=False, stubs=False):
    '''Regenerate generated modules with new genrator
    '''
    if services is stubs is False:
        return
    cp = ConfigParser()
    cp.read(CONFIG)
    for s,f in cp.items(GLOBUS_SECTION):
        print "Generate: ",s,f
        wsdl = GetWSDL(f)
        if services: 
            WriteServiceFile(wsdl)
        if stubs: 
            BetaWriteStubs(wsdl)
            BetaWriteStubsTypes(wsdl)
    for s,f in cp.items(LOCAL_SECTION):
        print "Generate: ",s,f
        wsdl = GetWSDL(f, isGlobus=False)
        if services: 
            WriteServiceFile(wsdl)
        if stubs: 
            BetaWriteStubs(wsdl)
            BetaWriteStubsTypes(wsdl)
    

class ServerCallbackDescription:
    method_prefix = 'wsa'
    ID1 = '    '
    ID2 = ID1 * 2
    ID3 = ID1 * 3
    ID4 = ID1 * 4
    base_class = 'WSResource'
    base_class_mod = 'pyGridWare.wsrf.resource'
    stub_class_dir = 'pyGridWare.generated.stubs'

    def __init__(self):
        self.imports  = None
        self.service  = ''
        self.classdef = ''
        self.initdef  = ''
        self.location = ''
        self.methods  = []
        self.action_in = {}
        self.action_out = {}

    def getMethodName(self, method):
        return '%s_%s' %(self.method_prefix, method)

    def getStubName(self):
        return '%s_server' % self.service

    def fromWsdl(self, ws):
        wsm = WriteServiceModule(ws)
        self.service = wsm.get_module_names()[1]
        wsm = None
        ws = ZSI.wsdlInterface.ZSIWsdlAdapter(ws)
        self.setImports()
        for service in ws.getServicesList():
            for port in service.getPortList():
                # fetch the service location
                for e in port.getExtensions():
                    soapAddress = None
                    if isinstance(e, ZSI.wsdlInterface.ZSISoapAddressAdapter):
                        soapAddress = e
                    if soapAddress:
                        self.location = soapAddress.getLocation()

                # generate methods
                for op in port.getBinding().getPortType().getOperationList():
                    self.generateMethod(op, port)

        self.setClassDef(ws)
        self.setInitDef()

    def write(self, fd=sys.stdout):
        module_contents = \
            string.join([self.imports, self.classdef, self.initdef, string.join(self.methods, '\n')], '\n') + '\n'
        fd.write(module_contents)

    def setImports(self):
        i  = 'from %s.%s import *' %(self.stub_class_dir, self.service)
        i += '\nfrom %s import %s' %(self.base_class_mod, self.base_class)
        self.imports = i

    def setClassDef(self, ws):
        c  = '\nclass %s(%s):' %(ws.getName(), self.base_class)

        # Set up soapAction dict
        sa = '\n%ssoapAction = {' % ServerCallbackDescription.ID1
        for method,value in self.action_in.items():
            sa += "\n%s'%s': '%s'," % (ServerCallbackDescription.ID2, value, self.getMethodName(method))
        sa += '\n%s}' % ServerCallbackDescription.ID2

        # Set up wsAction dict
        wsa = '\n%swsAction = {' % ServerCallbackDescription.ID1
        for method,value in self.action_out.items():
            wsa += "\n%s'%s': '%s'," % (ServerCallbackDescription.ID2, self.getMethodName(method), value)
        wsa += '\n%s}' % ServerCallbackDescription.ID2

        self.classdef = c + sa + wsa

    def setInitDef(self):
        uri = urlparse.urlparse( self.location )[2]
        d  = "\n%sdef __init__(self, post='%s', **kw):" % (ServerCallbackDescription.ID1, uri)
        d += '\n%s%s.__init__(self, post)' %(ServerCallbackDescription.ID2, self.base_class)
        self.initdef  = d

    def generateMethod(self, op, port):
        # generate wsActions while we're here
        operation_name = op.getName()
        # Grab portType from Adapter
        portType = port.getBinding().getPortType()._portType
        operation = portType.operations.get(operation_name)
        wsaction_in = wsaction_out = None
        if operation.input is not None:
            wsaction_in = operation.getInputAction()
        if operation.output is not None:
            wsaction_out = operation.getOutputAction()
        soap_operation = port.getBinding().getOperationDict().get(operation_name).getSoapOperation()
        if ALLOW_SLOPPY_WSDL is False:
            soap_operation = port.getBinding().getOperationDict().get(operation_name).getSoapOperation()
            soapaction = soap_operation.getAction()
            if soapaction and wsaction_in and soapaction != wsaction_in:
                raise RuntimeError, 'Bad WSDL operation(%s), soapAction(%s) != wsAction(%s)'\
                    %(operation_name, soapaction, wsaction_in)

        self.action_in[operation_name] = wsaction_in
        self.action_out[operation_name] = wsaction_out
 
        # now take care of the method
        inMessage = op.getInput().getMessage()
        if inMessage is not None:
            o  = '\n%sdef %s(self, ps, address):' % (ServerCallbackDescription.ID1, self.getMethodName(op.getName()))
            o += '\n%s# input vals in request object' % ServerCallbackDescription.ID2
            o += '\n%sself.request = ps.Parse(%s.typecode)' \
                      % ( ServerCallbackDescription.ID2, op.getInput().getMessage().getName() )
        else:
            o  = '\n%sdef %s(self):' % (ServerCallbackDescription.ID1, self.getMethodName(op.getName()))
            o += '\n%s# NO input' % ServerCallbackDescription.ID2
        
        outMessage = op.getOutput().getMessage()
        if outMessage is not None:
            o += '\n\n%s# assign return values to response object' % ServerCallbackDescription.ID2
            o += '\n%sresponse = %s()' % ( ServerCallbackDescription.ID2, outMessage.getName() )
            o += '\n\n%sreturn response' % ServerCallbackDescription.ID2
        else:
            o += '\n\n%s# NO output' % ServerCallbackDescription.ID2
            o += '\n\n%sreturn None' % ServerCallbackDescription.ID2

        self.methods.append(o)


class PyGridWareDist(Distribution): 
    global_options = Distribution.global_options[:]
    global_options.append(('regenerate', None, "regenerate service files"))

    def __init__ (self, attrs=None):
        # JRB Hook to run generator, wish there was a cleaner way??
        global_args = attrs.get('script_args', [])
        for arg in global_args:
            if arg == '--regenerate':
                BetaRegenerateModules(services=True, stubs=True)
        Distribution.__init__(self, attrs)


def main():
    classifiers = """\
        Development Status :: 5 - Production/Stable
        Intended Audience :: Developers
        Intended Audience :: Science/Research
        License :: OSI Approved :: BSD License
        Programming Language :: Python
        Natural Language :: English
        Operating System :: Microsoft :: Windows
        Operating System :: POSIX
        """

    longDesc = """
        Python implementation of the WS-Resource Framework (WSRF).  
        pyGridWare has facilities for secure messaging including security 
        context negotiation, delegation, and GSS XML Signatures.
        """

    setup(name="pyGridWare", 
          version="1.0.2beta", 
          packages=["pyGridWare", "pyGridWare.addressing", "pyGridWare.security", "pyGridWare.security.gss", "pyGridWare.security.base",
          "pyGridWare.generated", "pyGridWare.generated.types", "pyGridWare.generated.stubs",
          "pyGridWare.generated.services", "pyGridWare.generated.properties", "pyGridWare.properties",
          "pyGridWare.resource", "pyGridWare.lifetime", "pyGridWare.resource", "pyGridWare.notification", 
          "pyGridWare.generated.resource", "pyGridWare.services", "pyGridWare.wsrf", "pyGridWare.utility"],
          description="Python WS-Resource Framework and GSS Security",
	  author="Joshua Boverhof", 
          author_email="JRBoverhof@lbl.gov",
          maintainer="Joshua Boverhof", 
          maintainer_email="JRBoverhof@lbl.gov",
	  url="http://www-itg.lbl.gov/gtg/pyGridWare/",
          classifiers = filter(None, classifiers.split("\n"))[:-1],
          long_description=longDesc,
          distclass=PyGridWareDist,
    )

if __name__ == '__main__': 
    main()
