#! /usr/bin/env python
# $Header: /cvsroot/pywebsvcs/zsi/ZSI/fault.py,v 1.8.2.7 2004/11/24 00:09:34 boverhof Exp $
'''Faults.
'''

from ZSI import _copyright, _children, _child_elements, \
        _textprotect, _stringtypes, _seqtypes, _Node, SoapWriter
from ZSI.wstools.Namespaces import SOAP, ZSI_SCHEMA_URI
from ZSI.wstools.c14n import Canonicalize
import traceback, cStringIO as StringIO

class Fault:
    '''SOAP Faults.
    '''

    Client = "SOAP-ENV:Client"
    Server = "SOAP-ENV:Server"
    MU     = "SOAP-ENV:MustUnderstand"

    def __init__(self, code, string,
                actor=None, detail=None, headerdetail=None):
        self.code, self.string, self.actor, self.detail, self.headerdetail = \
                code, string, actor, detail, headerdetail

    def _do_detail_dict(self, elt, detail):
        '''XXX Hack to get this to work, w/o too much work
        '''
        for k,v in detail.items():
           namespaceURI,localName = k
           celt = elt.createAppendElement(namespaceURI, localName)
           if type(v) is dict:
               self._do_detail_dict(celt, v)
           elif v:
               celt.createAppendTextNode(str(v))
              
    def _do_details(self, elt, header):
        if header:
            detail = self.headerdetail
            celt = elt.createAppendElement(ZSI_SCHEMA_URI, 'detail')
        else:
            detail = self.detail
            celt = elt.createAppendElement(ZSI_SCHEMA_URI, 'FaultDetail')
        if type(detail) in _stringtypes:
            celt = celt.createAppendElement(ZSI_SCHEMA_URI, 'string')
            celt.createAppendTextNode(detail)
        elif type(detail) is dict:
            self._do_detail_dict(celt, detail)
        else:
            detail,tracetext = detail
            celt = celt.createAppendElement(ZSI_SCHEMA_URI, 'string')
            celt.createAppendTextNode(detail)
            celt = celt.createAppendElement(ZSI_SCHEMA_URI, 'trace')
            celt.createAppendTextNode(tracetext)

    def DataForSOAPHeader(self, elt):
        if not self.headerdetail: return None
        # SOAP spec doesn't say how to encode header fault data.
        self._do_details(elt, True)

    #def serialize(self, sw):
    def serialize(self, elt, sw, pyobj):
        '''Serialize the object.'''
        celt = elt.createAppendElement(SOAP.ENV, 'Fault')
        code = celt.createAppendElement(None, 'faultcode')
        code.createAppendTextNode(self.code)
        fstring = celt.createAppendElement(None, 'faultstring')
        fstring.createAppendTextNode(self.string)
        
        if self.actor:
            actor = celt.createAppendElement(SOAP.ENV, 'faultactor')
            actor.createAppendTextNode(self.actor)
        if self.detail: self._do_details(elt, False)

    def AsSOAP(self, output=None, **kw):
        if output is None:
            s = StringIO.StringIO()
            output = s
        else:
            s = None
        #mykw = { 'header': self.DataForSOAPHeader() }
        #if kw: mykw.update(kw)
        #sw = SoapWriter(output, **mykw)
        sw = SoapWriter()
        #self.serialize(sw)
        sw.serialize(self, typecode=self)
        header_elt = sw.getSOAPHeader()
        self.DataForSOAPHeader(header_elt)
        #if s: 
        #   return s.getvalue()
        #lse:
        #   return None
        msg = str(sw)
        return msg
    AsSoap = AsSOAP

def FaultFromNotUnderstood(uri, localname, actor=None):
    #elt = '''<ZSI:URIFaultDetail>
    #<ZSI:URI>%s</ZSI:URI>
    #<ZSI:localname>%s</ZSI:localname>
    #</ZSI:URIFaultDetail>
    #''' % (uri, localname)
    elt = {(ZSI_SCHEMA_URI,'URIFaultDetail'):\
        { (ZSI_SCHEMA_URI,'URI'):uri, (ZSI_SCHEMA_URI,'localname'):localname }
    }
    detail, headerdetail = None, elt
    return Fault(Fault.MU, 'SOAP mustUnderstand not understood',
                actor, detail, headerdetail)

def FaultFromActor(uri, actor=None):
    elt = '''<ZSI:ActorFaultDetail>
<ZSI:URI>%s</ZSI:URI>
</ZSI:ActorFaultDetail>
''' % uri
    detail, headerdetail = None, elt
    return Fault(Fault.Client, 'Cannot process specified actor',
                actor, detail, headerdetail)

def FaultFromZSIException(ex, actor=None):
    '''Return a Fault object created from a ZSI exception object.
    '''
    mystr = getattr(ex, 'str', None) or str(ex)
    mytrace = getattr(ex, 'trace', None) or ''
    elt = '''<ZSI:ParseFaultDetail>
<ZSI:string>%s</ZSI:string>
<ZSI:trace>%s</ZSI:trace>
</ZSI:ParseFaultDetail>
''' % (_textprotect(mystr), _textprotect(mytrace))
    if getattr(ex, 'inheader', 0):
        detail, headerdetail = None, elt
    else:
        detail, headerdetail = elt, None
    return Fault(Fault.Client, 'Unparseable message',
                actor, detail, headerdetail)

def FaultFromException(ex, inheader, tb=None, actor=None):
    '''Return a Fault object created from a Python exception.
    <ZSI:FaultDetail>
      <ZSI:string>%s</ZSI:string>
      <ZSI:trace>%s</ZSI:trace>
    </ZSI:FaultDetail> % ( _textprotect(str(ex)), tracetext)
    '''
    tracetext = None
    if tb:
        try:
            lines = '\n'.join(['%s:%d:%s' % (name, line, func)
                        for name, line, func, text in traceback.extract_tb(tb)])
            tracetext = _textprotect(lines)
        except:
            raise
    detailtext = "%s"%_textprotect(str(ex))
    headerdetail = None
    if inheader:
        detail, headerdetail = None, detailtext
    elif tracetext is None:
        detail = detailtext
    else:
        detail = (detailtext,tracetext)

    return Fault(Fault.Server, 'Processing Failure',
                actor, detail, headerdetail)

def FaultFromFaultMessage(ps):
    '''Parse the message as a fault.
    '''
    d = { 'faultcode': None, 'faultstring': None, 'faultactor': None,
        'detail': None, }
    for elt in _child_elements(ps.body_root):
        n = elt.localName
        if n == 'detail':
            d['detail'] = _child_elements(elt)
        if n in [ 'faultcode', 'faultstring', 'faultactor' ]:
            d[n] = ''.join([E.nodeValue for E in _children(elt)
                            if E.nodeType 
                            in [ _Node.TEXT_NODE, _Node.CDATA_SECTION_NODE ]])
    return Fault(d['faultcode'], d['faultstring'],
                d['faultactor'], d['detail'])

if __name__ == '__main__': print _copyright
