############################################################################
# Monte M. Goode, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################

# main generator engine for new generation generator

# $Id: wsdl2python.py,v 1.1.2.12 2004/11/24 00:09:35 boverhof Exp $

import sys

from ZSI.wstools.Utility import Base, SplitQName
from ZSI.wstools import WSDLTools
from ZSI.wstools.WSDLTools import SoapAddressBinding,\
    SoapBodyBinding, SoapBinding,MimeContentBinding,\
    HttpUrlEncodedBinding
from ZSI.wstools.XMLSchema import ElementDeclaration
from ZSI.typeinterpreter import BaseTypeInterpreter
from ZSI.generate import WsdlGeneratorError, Wsdl2PythonError


from containers import *
from ZSI.generate import utility


# change to a singleton class when we settle on a pattern

class Config:
    
    readerclass = None
    writerclass = None
    serviceImports = []
    
    def clear():
        Config.readerclass = None
        Config.writerclass = None
        Config.serviceImports = []
        
    clear = staticmethod(clear)

class WriteServiceModule(Base):
    """top level driver class invoked by wsd2py"""
    def __init__(self, wsdl, addressing=False, notification=False,
                 configParser = None):
        Base.__init__(self)
        self.logger.debug('WriteServiceModule ctor()')
        
        self._wsdl = wsdl
        self._addressing = addressing
        self._notification = notification
        self._configParser = configParser

        self.nsad = utility.NamespaceAliasDict()
        self.usedNamespaces = {}
        self.modPath = None

        self.services = []
        
    def write(self, schemaOnly=False):
        f_types, f_services = self.get_module_names()

        hasSchema = len(self._wsdl.types)
        if hasSchema:
            fd = open(f_types + '.py', 'w+')
            self.write_service_types(f_types, fd)
            fd.close()

        if schemaOnly:
            return

        fd = open(f_services + '.py', 'w+')
        self.write_services(f_types, f_services, fd, hasSchema)
        fd.close()

    def set_module_path(self, path):
        self.modPath = path

    def get_module_names(self):
        name = self._wsdl.name or self._wsdl.services[0].name
        if not name:
            raise WsdlGeneratorError, 'could not determine a service name'
        
        name = utility.nonColonizedName_to_moduleName(name)
        name = SplitQName(name)[1]
        
        f_types = '%s_services_types' % name
        f_services = '%s_services' % name

        if self.modPath:
            f_types = "%s/%s" % (self.modPath, f_types)
            f_services = "%s/%s" % (self.modPath, f_services)
        
        return f_types, f_services

    def write_service_types(self, f_types, fd):
        self.logger.debug('write_service_types()')

        header = '%s \n# %s.py \n# generated by %s\n%s\n\n'\
                  %('#'*50, f_types, self.__module__, '#'*50)
        fd.write(header)

        print >>fd, TypesHeaderContainer()

        self.gatherNamespaces()

        for v in self.usedNamespaces.values():
            sd = SchemaDescription()
            sd.fromWsdl(v)
            sd.write(fd)

    def gatherNamespaces(self):
        
        self.usedNamespaces = {}

        # gather up all the namespaces and their imports

        for schema in self._wsdl.types.values():
            self.logger.debug('Processing %s' % schema.getTargetNamespace())
            self.usedNamespaces[schema.getTargetNamespace()] = schema
            self.logger.debug('Imports: %s' % schema._imported_schemas)
            if schema._imported_schemas:
                self.getImports(schema._imported_schemas)

        # register namespaces with ns alias dict object
        
        for k,v in self.usedNamespaces.items():
            self.logger.debug('adding %s to ns alias dict' % k)
            self.nsad.add(k)
        
            
    def getImports(self, nsdict):
        self.logger.debug('getImports()')
        for k,v in nsdict.items():
            self.logger.debug('nsdict key(%s)' % k)
            self.usedNamespaces[k] = v
            if v._imported_schemas:
                self.getImports(v._imported_schemas)

    def write_services(self, f_types, f_services, fd, hasSchema):
        self.logger.debug('write_services()')
        
        header = '%s \n# %s.py \n# generated by %s\n%s\n\n'\
                  %('#'*50, f_services, self.__module__, '#'*50)
        fd.write(header)
        self.services = []
        for service in self._wsdl.services:
            sd = ServiceDescription(self._addressing)
            if hasSchema:
                sd.typesFile = f_types
            sd.fromWsdl(service)
            sd.write(fd)
            self.services.append(sd)
            
class ServiceDescription(Base):
    """client interface - locator, port, etc classes"""
    def __init__(self, addressing=False):
        Base.__init__(self)
        self.logger.debug('ServiceDescription ctor')

        self.typesFile = None
        self.wsAddressing = addressing

        self.nsad = utility.NamespaceAliasDict()

        self.imports   = ServiceHeaderContainer()
        self.interface = ServiceInterfaceContainer()
        self.locator   = ServiceLocatorContainer()
        self.methods   = []
        self.messages  = []

    def fromWsdl(self, service):

        # import statements
        self.imports.types = self.typesFile
        
        if Config.serviceImports:
            self.imports.extras += Config.serviceImports
        
        self.interface.serviceName = service.name
        self.locator.serviceName = service.name

        for p in service.ports:

            # locator and interface classes
            
            ptName = p.getBinding().getPortType().name
            bName =  p.getBinding().name
            
            try:
                addrBind = p.getAddressBinding()
            except WSDLTools.WSDLError, ex:
                self.logger.warning('Skip port(%s), missing address binding' %p.name)
                continue
            if isinstance(addrBind, SoapAddressBinding) is False:
                self.logger.warning('Skip port(%s), not a SOAP-1.1 address binding' %p.name)
                continue

            sAddress = addrBind.location
            self.interface.addPortInfo(ptName)
            self.locator.addPortInfo(ptName, bName, sAddress)
            
            # operation class

            rProp = p.getBinding().getPortType().getResourceProperties()

            mKlass =  ServiceOperationsClassContainer()
            
            mKlass.rProp = rProp
            mKlass.wsA   = self.wsAddressing
            mKlass.bName = bName

            if Config.readerclass:
                mKlass.reader = Config.readerclass
            if Config.writerclass:
                mKlass.writer = Config.writerclass

            # operation methods

            for op in p.getBinding().getPortType().operations:

                # use, transport, style and etc...
                
                for bop in p.getBinding().operations:
                    if bop.name == op.name:
                        break
                else:
                    raise Wsdl2PythonError, 'no bop name(%s)' % op.name

                sob = p.getBinding().findBinding(SoapBinding)
                if not sob:
                    raise Wsdl2PythonError,\
                        'port(%s) operation(%s) missing SoapBinding'\
                        %(p.name, bop.name)

                transport = sob.transport
                style = sob.style
                soapOpBinding = bop.findBinding(WSDLTools.SoapOperationBinding)
                if not soapOpBinding:
                    self.logger.warning(\
                        'Skip port(%s) operation(%s) no SOAP Binding Operation'\
                        %(p.name, bop.name))
                    continue

                soapAction = soapOpBinding.soapAction
                if bop.input is not None:
                    
                    if not bop.input.findBinding(SoapBodyBinding):
                        
                        if bop.input.findBinding(HttpUrlEncodedBinding):
                            self.logger.warning(\
                                'Skip port(%s) operation(%s) HTTP Binding ops not supported'\
                                %(p.name, bop.name))
                            continue
                        elif bop.input.findBinding(MimeContentBinding):
                            self.logger.warning(\
                                'Skip port(%s) operation(%s) MIME Bindings opt not supported'\
                                %(p.name, bop.name))
                            continue
                        raise Wsdl2PythonError, \
                            'Missing SOAP-1.1 body binding for port(%s)' %p.name
                    else:
                        soapBodyBind = bop.input.findBinding(SoapBodyBinding)
                    
                    encodingStyle = soapBodyBind.encodingStyle
                    namespace = soapBodyBind.namespace
                    use = soapBodyBind.use

                # build the operation object

                oc = ServiceOperationContainer()
                oc.name = op.name
                oc.wsA  = self.wsAddressing
                oc.soapaction = soapAction
                
                if op.input:
                    oc.inputName  = op.getInputMessage().name
                    oc.inputSimpleType = self.isSimpleElementDeclaration(op)
                    oc.inputAction = op.getInputAction()
                
                if op.output:
                    oc.outputName = op.getOutputMessage().name
                    oc.outputSimpleType = \
                                        self.isSimpleElementDeclaration(op,
                                                                        False)
                    oc.outputAction = op.getOutputAction()

                mKlass.addOperation(oc)

                # Only care about request response, and request only messaging patterns.
                # For response only you need a Notification Consumer.
                rpc,literal = (style=='rpc'),(use=='literal')
                if op.input:
                    self.messages.append(\
                        MessageWriter(op, bop, input=True, rpc=rpc, literal=literal, simple=oc.inputSimpleType),
                        )
                    if op.output:
                        self.messages.append(\
                            MessageWriter(op, bop, input=False, rpc=rpc, literal=literal, simple=oc.outputSimpleType),
                            )
            self.methods.append(mKlass)

    def write(self, fd):

        print >>fd, self.imports
        print >>fd, self.interface
        print >>fd, self.locator
        for m in self.methods:
            print >>fd, m
        for m in self.messages:
            print >>fd, m

    def isSimpleElementDeclaration(self, op, input=True):
        '''If message consists of one part with an element attribute,
        and this element is a simpleType return a string representing 
        the python type, else return None.

        op    -- WSDLTools operation instance.
        input -- if True input, else output
        '''
        message = None
        if input is True:
            message = op.getInputMessage()
        else:
            message = op.getOutputMessage()

        if len( message.parts ) == 1:
            # bulletproofing against the phantom collection issue.
            wsdl = message.getWSDL()
            types = wsdl.types
            #part = message.parts.data.values()[0]
            part = message.parts[0]
            if not part: raise RuntimeError, 'what??'
            #types = part.parent().parent().parent().parent().types

            # Do we have element or type??
            if part.element:
               nsuri,name = part.element

               if types.has_key(nsuri) and types[nsuri].elements.has_key(name):
                    e = types[nsuri].elements[name]

                    if isinstance(e, ElementDeclaration) is True and e.getAttribute('type'):
                        typ = e.getAttribute('type')
                        bt = BaseTypeInterpreter()
                        ptype = bt.get_pythontype(typ[1], typ[0])
                        return ptype

        return None


class MessageWriter(Base):
    def __init__(self, operation, boperation, input=False, rpc=False, literal=True, simple=None):
        '''Representation of a WSDL Message and associated WSDL Binding.
        operation --
        boperation --
        input --
        rpc --
        literal --
        simple --
        '''
        Base.__init__(self)
        self.logger.debug('MessageWriter ctor()')
        self.nsad = utility.NamespaceAliasDict()
        self.content = None
        self.operation = operation
        self.bop = boperation
        self.input = input
        self.setUp(rpc, literal, simple)
       
    def setUp(self, rpc, literal, simple): 
        if self.bop.name != self.operation.name:
            raise Wsdl2PythonError, 'Mismatch between binding operation(%s) and operation(%s)'\
                %(self.bop.name, self.operation.name)

        if literal is False:
            raise Wsdl2PythonError, 'Only support doc/literal and rpc/literal.'
        if rpc is True:
            self.fromMessageRPCLiteral()
        else:
            self.fromMessageDocumentLiteral(simple)

    def __str__(self):
        if not self.content:
            raise Wsdl2PythonError, 'no message container installed!'
        return self.content.getvalue()
        
    def fromMessageDocumentLiteral(self, simple):
        '''process document/literal part.
        Set up instance variable content.
        '''
        self.content = ServiceDocumentLiteralMessageContainer(simple)
        if self.input is True: 
            message = self.operation.getInputMessage()
        else:
            message = self.operation.getOutputMessage()

        # using underlying data structure to avoid phantom problem.
        parts = message.parts.data.values()
        if len(parts) > 1:
            raise Wsdl2PythonError, 'not suporting multi part doc/lit msgs'
        elif len(parts) == 0:
            raise Wsdl2PythonError, 'must specify part for doc/lit msg'

        p = parts.pop()
        if p.element:
            self.content.ns = p.element[0]
            self.content.pName = p.element[1]
            self.content.mName = message.name
        elif p.type:
            raise  Wsdl2PythonError, 'no doc/lit suport for <part type>'

    def fromMessageRPCLiteral(self):
        '''process rpc/literal part(s)
        Set up instance variable content.
        '''
        self.content = ServiceRPCLiteralMessageContainer(self.operation, self.bop, self.input)


class SchemaDescription(Base):
    """generates classes for defs and decs in the schema instance"""
    def __init__(self):
        Base.__init__(self)
        self.logger.debug('SchemaDescription ctor()')

        self.nsad = utility.NamespaceAliasDict()

        self.classHead = NamespaceClassHeaderContainer()
        self.classFoot = NamespaceClassFooterContainer()

        self.types = []

    def fromWsdl(self, schema):

        ns = schema.getTargetNamespace()

        self.classHead.ns = self.classFoot.ns = ns

        self.generate(schema.types)
        self.generate(schema.elements)

    def write(self, fd):

        print >>fd, self.classHead

        for t in self.types:
            print >>fd, t

        print >>fd, self.classFoot

    def generate(self, types):

        for tp in types:
            t = TypeWriter()
            t.fromType(tp)
            self.types.append(t)

class TypeWriter(Base):
    """contains/generates a single definition/declaration"""
    def __init__(self):
        Base.__init__(self)
        self.logger.debug('TypeWriter ctor()')
        self.content = None
        self.bti = BaseTypeInterpreter()
        
    def fromType(self, tp):

        if tp.isDefinition():
            self.fromDefinition(tp)
        elif tp.isElement():
            self.fromElement(tp)
        else:
            raise Wsdl2PythonError, 'Not element nor def?'

    def fromElement(self, tp):

        qName = tp.getAttribute('type')

        local = False
        globl = False
        
        if not qName:
            etp = tp.content
            local = True
        else:
            etp = tp.getTypeDefinition('type')
            globl = True
            

        if etp is None:
            if globl:
                self._fromElementSimpleType(tp, qName)
            else:
                # empty element declaration.
                # not exactly, but it'll do the right thing...
                self._fromElementWithLocalDef(tp)
        elif etp.isSimple(): 
            if globl:
                self._fromElementGlobalDef(tp, etp)
            else:
                raise Wsdl2PythonError, 'local simple?'
        elif etp.isComplex():
            if local:
                self._fromElementWithLocalDef(tp)
            elif globl:
                self._fromElementGlobalDef(tp, etp)
            else:
                raise Wsdl2PythonError, 'neither global or local?'
        else:
            raise Wsdl2PythonError, "What the Hell is this?"

        
    def fromDefinition(self, tp):

        content = tp.content

        simple  = False
        complex = False
        
        if tp.isSimple():
            simple = True
        elif tp.isComplex():
            complex = True
        else:
            raise Wsdl2PythonError, "not simple or complex?"

        # an empty complex type <complexType name="foo"></complexType>
        if tp.isComplex() and not content:
            self._fromModelGroup(tp, content)
            return

        if content.isSimple():
            if complex:
                self._fromSimpleContent(tp, content)
            else:
                raise Wsdl2PythonError, 'simple type, simple content?'
        elif content.isComplex():
            if complex:
                self._fromComplexContent(tp, content)
            else:
                raise Wsdl2PythonError, "simple type, complex content?"
        elif content.isRestriction():
            if simple:
                self._fromSimpleType(tp)
        elif content.isModelGroup():
            if complex:
                self._fromModelGroup(tp, content)
            else:
                raise Wsdl2PythonError, "simple type, model group?"
        elif content.isUnion():
            self._fromUnion(tp)
        elif content.isList():
            self._fromList(tp)
        else:
            raise Wsdl2PythonError, "content type unknown: %s" % content

    def _fromElementSimpleType(self, tp, qName):

        self.content = ElementSimpleTypeContainer()
        self.content.name = tp.getAttribute('name')
        self.content.ns = tp.getTargetNamespace()
        self.content.sKlass = self.bti.get_typeclass(qName[1], qName[0])

    def _fromElementWithLocalDef(self, tp):

        self.content = ElementLocalDefContainer()
        self.content.name = tp.getAttribute('name')
        self.content.ns = tp.getTargetNamespace()

        if hasattr(tp.content, 'content') and tp.content.content and \
               hasattr(tp.content.content, 'content') and \
               tp.content.content.content:
            self.content.mgContent = tp.content.content.content
        else:
            # void -> <complexType/>
            self.content.mgContent = ()

    def _fromElementGlobalDef(self, tp, etp):

        self.content = ElementGlobalDefContainer()
        self.content.name = tp.getAttribute('name')
        self.content.ns = tp.getTargetNamespace()
        self.content.sKlass = etp.getAttribute('name')
        self.content.sKlassNS = etp.getTargetNamespace()

    def _fromComplexContent(self, tp, content):

        if hasattr(tp.content, 'derivation') and tp.content.derivation:

            self.content = DerivedTypeDefContainer()
            self.content.name = tp.getAttribute('name')
            self.content.ns = tp.getTargetNamespace()

            base = tp.content.derivation.getTypeDefinition('base')

            if not base:
                if tp.content.derivation.attributes.has_key('base') and \
                       tp.content.derivation.attributes['base'] == \
                       (u'http://schemas.xmlsoap.org/soap/encoding/',u'Array'):
                    raise Wsdl2PythonError, \
                          'soap-enc arrays not supported!'
                else:
                    raise Wsdl2PythonError, \
                          'could not determine derivation base!'

            self.content.sKlass = base.getAttribute('name')
            self.content.sKlassNS = base.getTargetNamespace()

            if tp.content.derivation.isRestriction():
                self.content.restriction = True
                self.content.extension = False
                # data for attributes
                if hasattr(tp, 'getAttributeContent'):
                    self.content.attributeContent = tp.getAttributeContent()
            elif tp.content.derivation.isExtension():
                self.content.restriction = False
                self.content.extension = True
                # data for attributes
                attrs = []
                if hasattr(tp, 'getAttributeContent') and \
                   tp.getAttributeContent() != None:
                    attrs += tp.getAttributeContent()
                if hasattr(tp.content.derivation, 'getAttributeContent') and \
                   tp.content.derivation.getAttributeContent():
                    attrs += tp.content.derivation.getAttributeContent()
                if attrs:
                    self.content.attributeContent = attrs
                    self.content.extType = tp.content.derivation
            else:
                raise Wsdl2PythonError, 'neither restriction nor extension?'

            if content.derivation.content:
                self.content.mgContent = content.derivation.content.content
            else:
                self.content.mgContent = ()

        else:
            raise Wsdl2PythonError, \
                  'need to handle new form of complex content'

    def _fromModelGroup(self, tp, content):

        self.content = ModelGroupContainer()
        self.content.name = tp.getAttribute('name')
        self.content.ns = tp.getTargetNamespace()

        if tp.getAttribute('mixed'):
            self.content.mixed = True

        if content and hasattr(content, 'content') and content.content:
            self.content.mgContent = content.content
            if content.getAttribute('mixed'):
                self.content.mixed = True
        else:
            self.content.mgContent = ()

        if hasattr(tp, 'getAttributeContent'):
            self.content.attributeContent = tp.getAttributeContent()
            
    def _fromSimpleType(self, tp):
        '''Create a simpleType definition
        '''
        pyclass = None
        if hasattr(tp, 'content') and tp.content.getAttribute('base'):
            # the "usual" option
            base = tp.content.getAttribute('base')
            pyclass = self.bti.get_typeclass( base[1], base[0] )
        elif hasattr(tp, 'content') and hasattr(tp.content, 'content') \
             and tp.content.content:
            # perhaps buried in an anonymous sub type?
            base = None
            for c in tp.content.content:
                if hasattr(c, 'content') and c.content.getAttribute('base'):
                    base = c.content.getAttribute('base')
                    break
            if base:
                pyclass = self.bti.get_typeclass(base[1], base[0])
            else:
                raise Wsdl2PythonError, \
                      'could not get restriction base in subtypes!'
        else:
            raise Wsdl2PythonError, 'could not determine restriction base!'

        name = tp.getAttribute('name')
        tns = tp.getTargetNamespace()
        self.content = SimpleTypeContainer()
        self.content.setUp(name, tns, pyclass)

    def _fromSimpleContent(self, tp, content):

        base = None

        if hasattr(tp, 'content') and tp.content and \
           hasattr(tp.content, 'derivation'):
            base = tp.content.derivation.getAttribute('base')

        if base:
            tc = self.bti.get_typeclass( base[1],
                                         base[0] )
            if tc:
                # then we're extending a primitive type
                self.content = SimpleContentContainer()
                self.content.name = tp.getAttribute('name')
                self.content.ns = tp.getTargetNamespace()
                self.content.sKlass = tc
            else:
                # we're extending a defined type, so swap in a different
                # container.
                self.content = DerivedSimpleContentContainer()
                self.content.name = tp.getAttribute('name')
                self.content.ns = tp.getTargetNamespace()
                self.content.sKlass   = base[1]
                self.content.sKlassNS = base[0]
        else:
            raise Wsdl2PythonError, 'could not determine simple content base!'

        if hasattr(tp, 'content') and tp.content and \
           hasattr(tp.content, 'derivation') and tp.content.derivation:
            if hasattr(tp.content.derivation, 'getAttributeContent'):
                atts = tp.content.derivation.getAttributeContent()
                self.content.attributeContent = atts
        
    def _fromUnion(self, tp):

        # placeholder pending further work
        self.content = UnionOrListContainer()
        self.content.name = tp.getAttribute('name')
        self.content.ns = tp.getTargetNamespace()
        self.content.sKlass = 'ZSI.TC.String'

    def _fromList(self, tp):

        # placeholder pending further work
        self.content = UnionOrListContainer()
        self.content.name = tp.getAttribute('name')
        self.content.ns = tp.getTargetNamespace()
        self.content.sKlass = 'ZSI.TC.String'
        

    def __str__(self):
        if not self.content:
            raise Wsdl2PythonError, 'no typecode container installed!'

        out = self.content.getvalue()

        # handle subtypes if they exist
        if self.content.subTypes:
            parentName = self.content.getClassName()
            parentNs   = self.content.getNSAlias()
            for s in self.content.subTypes:
                tw = self.__class__()
                tw.fromType(s)
                out += '\n\n'
                # register parent class w/container
                if self.content.parentClass:
                    tw.content.parentClass = '%s.%s' \
                                             % (self.content.parentClass,
                                                parentName)
                else:
                    tw.content.parentClass = '%s.%s' %(parentNs, parentName)
                for l in tw.content.getvalue().split('\n'):
                    if l:
                        out += '%s%s\n' % (ID1, l)
                    else:
                        out += '\n'
                out += '\n\n'
                                    
        return out
        
