#! /usr/bin/env python
# $Header: /cvsroot/pywebsvcs/zsi/ZSI/writer.py,v 1.14.2.10 2004/11/08 19:44:48 boverhof Exp $
'''SOAP message serialization.
'''

from ZSI import _copyright, ZSI_SCHEMA_URI
from ZSI import _backtrace, _stringtypes, _seqtypes
from ZSI.TC import AnyElement, TypeCode
from ZSI.wstools.Utility import MessageInterface, ElementProxy
from ZSI.wstools.Namespaces import XMLNS, SOAP, SCHEMA
from ZSI.wstools.c14n import Canonicalize
import types

_standard_ns = [ ('xml', XMLNS.XML), ('xmlns', XMLNS.BASE) ]

_reserved_ns = {
        'SOAP-ENV': SOAP.ENV,
        'SOAP-ENC': SOAP.ENC,
        'ZSI': ZSI_SCHEMA_URI,
        'xsd': SCHEMA.BASE,
        'xsi': SCHEMA.BASE + '-instance',
}

class SoapWriter:
    '''SOAP output formatter.
       Instance Data:
           out --
           memo --
           envelope --
           encoding -- 
           header -- list of DOM's or TypeCodes to serialize in header.
    '''

    def __init__(self, envelope=True, encoding=SOAP.ENC, header=True, 
    nsdict={}, outputclass=None, **kw):
        '''Initialize.
        '''
        if outputclass is None:
            outputclass=ElementProxy
        if not issubclass(outputclass, MessageInterface):
            raise TypeError, 'outputclass must subclass MessageInterface'

        self.dom, self.memo, self.nsdict= \
            outputclass(self), [], nsdict
        self.envelope = envelope
        self.encoding = encoding
        self.header = header
        self.body = None

    def __str__(self):
        return str(self.dom)

    def getSOAPHeader(self):
        if self.header is True or self.header is False:
            return None
        return self.header

    def serialize_header(self, hpyobjs, htypecode, **kw):
        '''Serialize a Python object in SOAP-ENV:Header, make
        sure everything in Header unique (no #href).
        '''
        kw['unique'] = True
        soap_env = _reserved_ns['SOAP-ENV']
        header = self.dom.getElement(soap_env, 'Header')
        if type(hpyobjs) not in _seqtypes:
           hpyobjs = (hpyobjs,)
        for hpyobj in hpyobjs:
            helt = htypecode.serialize(header, self, hpyobj, **kw)
            #Handle SOAP attributes in SoapWriter.
            if hasattr(hpyobj, 'attributes'):
                if hpyobj.attributes.has_key((SOAP.ENV,'mustUnderstand')):
                    value = hpyobj.attributes.has_key((SOAP.ENV,'mustUnderstand'))
                    helt.setAttributeNS(SOAP.ENV, 'mustUnderstand', value)


    def serialize(self, pyobj, typecode=None, root=None, header_pyobjs={}, header_typecodes=(), **kw):
        '''Serialize a Python object to the output stream.
           pyobj -- python instance to serialize in body.
           typecode -- typecode describing body 
           root -- SOAP-ENC:root
           header_pyobjs -- dictionary of header instances
           header_typecodes -- list of header typecodes
        '''
        self.body = None
        if self.envelope: 
            soap_env = _reserved_ns['SOAP-ENV']
            self.dom.createDocument(soap_env, 'Envelope')
            for prefix, nsuri in _reserved_ns.items():
                self.dom.setNamespaceAttribute(prefix, nsuri)
            self.writeNSdict(self.nsdict)
            if self.encoding:
                self.dom.setAttributeNS(soap_env, 'encodingStyle', self.encoding)
            if self.header:
                header = self.dom.createAppendElement(soap_env, 'Header')
                #for htypecode,hpyobj in self.header:
                #    htypecode.serialize(header, self, hpyobj, **kw)
                for htypecode in header_typecodes:
                    nspname,pname = htypecode.nspname,htypecode.pname
                    hpyobjs = header_pyobjs.get((nspname,pname))
                    if hpyobjs is not None:
                        """
                        if type(hpyobjs) not in _seqtypes:
                           hpyobjs = (hpyobjs,)
                        if isinstance(htypecode, AnyElement):
                            hpyobjs = [header_pyobjs.get("http://counter.com", "CounterKey")]
                        for hpyobj in hpyobjs:
                            helt = htypecode.serialize(header, self, hpyobj, **kw)
                            #Handle SOAP attributes in SoapWriter.
                            if hasattr(hpyobj, 'attributes'):
                                if hpyobj.attributes.has_key((SOAP.ENV,'mustUnderstand')):
                                    value = hpyobj.attributes.has_key((SOAP.ENV,'mustUnderstand'))
                                    helt.setAttributeNS(SOAP.ENV, 'mustUnderstand', value)
                        """
                        self.serialize_header(hpyobjs, htypecode, **kw)

            self.body = self.dom.createAppendElement(soap_env, 'Body')

        if typecode is None: typecode = pyobj.__class__.typecode
        if TypeCode.typechecks and type(pyobj) == types.InstanceType and \
        not hasattr(typecode, 'pyclass'):
            pass
            # XXX XML ...
            #raise TypeError('Serializing Python object with other than Struct.')
        kw = kw.copy()
        if root in [ 0, 1 ]:
            kw['attrtext'] = ' SOAP-ENC:root="%d"' % root
        if self.body is None:
            typecode.serialize(self.dom, self, pyobj, **kw)
        else:
            typecode.serialize(self.body, self, pyobj, **kw)
        return self

    def writeNSdict(self, nsdict):
        '''Write a namespace dictionary, taking care to not clobber the
        standard (or reserved by us) prefixes.
        '''
        for k,v in nsdict.items():
            if (k,v) in _standard_ns: continue
            rv = _reserved_ns.get(k)
            if rv:
                if rv != v:
                    raise KeyError("Reserved namespace " + str((k,v)) + " used")
                continue
            if k:
                self.dom.setNamespaceAttribute(k, v)
            else:
                self.dom.setNamespaceAttribute('xmlns', v)


    def ReservedNS(self, prefix, uri):
        '''Is this namespace (prefix,uri) reserved by us?
        '''
        return _reserved_ns.get(prefix, uri) != uri

    def Known(self, obj):
        '''Seen this object (known by its id()?  Return 1 if so,
        otherwise add it to our memory and return 0.
        '''
        obj = id(obj)
        if obj in self.memo: return 1
        self.memo.append(obj)
        return 0

    def Forget(self, obj):
        '''Forget we've seen this object.
        '''
        obj = id(obj)
        try:
            self.memo.remove(obj)
        except ValueError:
            pass

    def Backtrace(self, elt):
        '''Return a human-readable "backtrace" from the document root to
        the specified element.
        '''
        return _backtrace(elt._getNode(), self.dom._getNode())



if __name__ == '__main__': print _copyright
